# proxmox-deploy is cli-based deployment tool for Proxmox
#
# Copyright (c) 2015 Nick Douma <n.douma@nekoconeko.nl>
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see http://www.gnu.org/licenses/.

from proxmoxdeploy.questions import QuestionGroup, OptionalQuestionGroup, \
    SpecificAnswerOptionalQuestionGroup, Question, BooleanQuestion, \
    EnumQuestion, NoAskQuestion, IntegerQuestion, MultipleAnswerQuestion
from jinja2 import Environment, PackageLoader, Template
from subprocess import Popen, PIPE
import locale
import os
import pytz

VALID_LOCALES = sorted(set(locale.locale_alias.values()))
VALID_KEYBOARD_LAYOUTS = [
    "af", "al", "am", "ara", "at", "az", "ba", "bd", "be", "bg", "br", "brai",
    "bt", "bw", "by", "ca", "cd", "ch", "cm", "cn", "cz", "de", "dk", "ee",
    "epo", "es", "et", "fi", "fo", "fr", "gb", "ge", "gh", "gn", "gr", "hr",
    "hu", "ie", "il", "in", "iq", "ir", "is", "it", "jp", "ke", "kg", "kh",
    "kr", "kz", "la", "latam", "lk", "lt", "lv", "ma", "mao", "md", "me", "mk",
    "ml", "mm", "mn", "mt", "mv", "nec_vndr/jp", "ng", "nl", "no", "np", "ph",
    "pk", "pl", "pt", "ro", "rs", "ru", "se", "si", "sk", "sn", "sy", "th",
    "tj", "tm", "tr", "tw", "tz", "ua", "us", "uz", "vn", "za"
]
VALID_TIMEZONES = sorted(pytz.common_timezones)
VALID_IMAGE_FORMATS = [".iso", ".img", ".qcow2", ".raw"]
VALID_COMPRESSION_FORMATS = [".xz", ".gz", ".bz2"]

try:
    agent = Popen(["ssh-add", "-L"], stdout=PIPE)
    if agent.wait() == 0:
        DEFAULT_SSH_KEYS = agent.stdout.read().rstrip().split("\n")
    del agent
except:
    DEFAULT_SSH_KEYS = None

QUESTIONS = QuestionGroup([
    ("_basic", QuestionGroup([
        ("name", Question("Hostname (a FQDN is recommended)")),
    ])),
    ("_languages", QuestionGroup([
        ("locale", EnumQuestion("Locale", default="en_US.UTF-8",
                                valid_answers=VALID_LOCALES)),
        ("timezone", EnumQuestion("Timezone", default="Europe/Amsterdam",
                                  valid_answers=VALID_TIMEZONES)),
        ("kb_layout", EnumQuestion("Keyboard layout", default="us",
                                   valid_answers=VALID_KEYBOARD_LAYOUTS)),
    ])),
    ("_security", QuestionGroup([
        ("ssh_pass_auth", NoAskQuestion("Allow SSH login using password",
                                        default=False)),
        ("ssh_root_keys", MultipleAnswerQuestion(
            "SSH Public key for root user", default=DEFAULT_SSH_KEYS)),
        ("apt_update", BooleanQuestion("Run apt-get update after rollout",
                                       default=True)),
        ("apt_upgrade", BooleanQuestion("Run apt-get upgrade after rollout",
                                        default=False))
    ])),
    ("_network", OptionalQuestionGroup([
        ("configure_network", NoAskQuestion(question=None, default=True)),
        ("vlan_id", IntegerQuestion("VLAN ID", default=1,
                                    min_value=1, max_value=4096)),
        ("network_device", Question(question="Network device to configure",
                                    default="eth0")),
        ("_static_network", SpecificAnswerOptionalQuestionGroup([
            ("ip_address", Question("IP Address")),
            ("subnet_mask", Question("Subnet Mask", default="255.255.255.0")),
            ("network_address", Question("Network Address")),
            ("broadcast_address", Question("Broadcast Address")),
            ("gateway_address", Question("Gateway Address")),
            ("dns_servers", Question("DNS Servers (space separated)"))
        ], optional_question=EnumQuestion(
            "Network type", default="dhcp", valid_answers=["static", "dhcp"]),
            specific_answer="static"
        ))
    ], optional_question=BooleanQuestion("Configure networking", default=True),
        negative_questions={"vlan_id": NoAskQuestion(question=None, default=1)}
    )),
    ("_misc", QuestionGroup([
        ("resize_rootfs", BooleanQuestion("Resize root filesystem",
                                          default=True)),
        ("packages", Question("Install extra packages (space separated))",
                              default="")),
        ("commands", Question(
            "Run commands after cloud init (space separated)", default=""))
    ]))
])


def ask_cloudinit_questions(cloud_images_dir):
    global QUESTIONS
    images = list_images(cloud_images_dir)
    if len(images) < 1:
        raise RuntimeError("Cloud images directory contains no valid images.")
    QUESTIONS['_basic']['image'] = EnumQuestion("What Cloud image to upload",
                                                valid_answers=images,
                                                default=images[0])
    QUESTIONS.ask_all()
    return QUESTIONS.flatten_answers()


def _generate_data(output_file, context, template_file, default_template):
    if not template_file:
        env = Environment(loader=PackageLoader("proxmoxdeploy.cloudinit"))
        template = env.get_template(default_template)
    else:
        template = Template(template_file.read())

    with open(output_file, "w") as output:
        output.write(template.render(context=context))


def list_images(_dir):
    """
    Walks the given directory recursively and list all usable images.
    """
    images = []
    for root, subdirs, files in os.walk(_dir):
        if subdirs:
            for subdir in subdirs:
                images = images + list_images(subdir)
        for _file in files:
            if os.path.splitext(_file)[1] in \
                    VALID_IMAGE_FORMATS + VALID_COMPRESSION_FORMATS:
                images.append(os.path.join(root, _file))
    return images


def generate_user_data(output_file, context, template_file=None):
    """
    Generates the user-data part of the "No Cloud" cloud-init approach.

    Parameters
    ----------
    output_file: str
        Filename where the user-data file will be created.
    context: dict
        Dict(-like) object where the required template variables can be looked
        up.
    template_file: file
        File to read the Jinja2 template to populate from. If not set, will
        load the default template. The file will be read to the end.
    """
    _generate_data(output_file, context, template_file, "user-data.j2")


def generate_meta_data(output_file, context, template_file=None):
    """
    Generates the meta-data part of the "No Cloud" cloud-init approach.

    Parameters
    ----------
    output_file: str
        Filename where the meta-data file will be created.
    context: dict
        Dict(-like) object where the required template variables can be looked
        up.
    template_file: file
        File to read the Jinja2 template to populate from. If not set, will
        load the default template. The file will be read to the end.
    """
    _generate_data(output_file, context, template_file, "meta-data.j2")



# proxmox-deploy is cli-based deployment tool for Proxmox
#
# Copyright (c) 2015 Nick Douma <n.douma@nekoconeko.nl>
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see http://www.gnu.org/licenses/.

from ..exceptions import CommandInvocationException
from .templates import generate_user_data, generate_meta_data
from distutils.spawn import find_executable
from shutil import rmtree
from subprocess import Popen, PIPE
import logging
import os
import shlex
import tempfile

GENISOIMAGE_COMMAND = find_executable("genisoimage")
if GENISOIMAGE_COMMAND is None:
    raise RuntimeError(
        "genisoimage command is missing, make sure it is installed.")

CLI_ECHO_COMMANDS = False
CLI_ECHO_COMMAND_MESSAGE = "  Running command: `{0}`"
CLI_COMMAND_PREFIX = ""

logger = logging.getLogger(__name__)


def call_cli(command, error_message=None, expected_return_code=0):
    """
    Calls the given CLI command and handles output. If the return code does
    not match the expected_return_code, a CommandInvocationException
    will be raised.

    Parameters
    ----------
    command: str
        Command to execute, will be parsed by shlex.
    error_message: str
        Custom error message to display if the return code does not match
        expected_return_code.
    expected_return_code: int
        Expected return code, raises a CommandInvocationException otherwise.

    Returns
    -------
    Tuple containing (stdout, stderr).
    """
    cmd = str(command.encode("utf-8").decode("ascii", "ignore"))
    if CLI_COMMAND_PREFIX:
        cmd = "{0} {1}".format(CLI_COMMAND_PREFIX, cmd)

    if CLI_ECHO_COMMANDS:
        print CLI_ECHO_COMMAND_MESSAGE.format(cmd)

    env = os.environ.copy()
    proc = Popen(shlex.split(cmd), stdout=PIPE, stderr=PIPE, env=env)
    stdout, stderr = proc.communicate()

    if proc.returncode != expected_return_code:
        if error_message:
            raise CommandInvocationException(
                message="{0}: {1}".format(error_message, stderr),
                stdout=stdout,
                stderr=stderr
            )
        else:
            raise CommandInvocationException(
                message="Failed to execute {0}: {1}".format(cmd, stderr),
                stdout=stdout,
                stderr=stderr
            )

    return (stdout, stderr)


def generate_seed_iso(context, output_file=None):
    """
    Calls genisofs to create an cloud-init compatible ISO file. This ISO file
    can be used to seed a cloud-init installation using the "No Cloud" approach
    (https://cloudinit.readthedocs.org/en/latest
    /topics/datasources.html#no-cloud).

    Parameters
    ----------
    output_file: str
        Filename where the ISO file will be created.
    context: dict
        Dict-like object to use as context for generating the user-data and
        meta-data files.
    """
    if not output_file:
        output_file = tempfile.mkstemp(prefix="cloudinit-seed-iso-",
                                       suffix=".iso")[1]

    temp_dir = tempfile.mkdtemp(prefix="cloudinit-seed-iso")
    logger.info("Generating cloud-init seed files at {0}".format(temp_dir))
    generate_user_data(os.path.join(temp_dir, "user-data"), context)
    generate_meta_data(os.path.join(temp_dir, "meta-data"), context)

    logger.info("Generating cloud-init seed ISO at {0}".format(output_file))
    call_cli("{0} -output '{1}' -volid cidata -joliet -rock '{2}'"
             .format(GENISOIMAGE_COMMAND, output_file, temp_dir))

    logger.info("Removing cloud-init temp files")
    rmtree(temp_dir)
    return output_file



# proxmox-deploy is cli-based deployment tool for Proxmox
#
# Copyright (c) 2015 Nick Douma <n.douma@nekoconeko.nl>
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see http://www.gnu.org/licenses/.

from .cloudinit.templates import VALID_IMAGE_FORMATS, VALID_COMPRESSION_FORMATS
from .exceptions import SSHCommandInvocationException
from .questions import QuestionGroup, IntegerQuestion, EnumQuestion, \
    NoAskQuestion
from openssh_wrapper import SSHError
import logging
import math
import os.path

CPU_FAMILIES = [
    "486", "athlon", "pentium", "pentium2", "pentium3", "coreduo", "core2duo",
    "kvm32", "kvm64", "qemu32", "qemu64", "phenom", "Conroe", "Penryn",
    "Nehalem", "Westmere", "SandyBridge", "IvyBridge", "Haswell", "Broadwell",
    "Opteron_G1", "Opteron_G2", "Opteron_G3", "Opteron_G4", "Opteron_G5",
    "host"
]

logger = logging.getLogger(__name__)


def ask_proxmox_questions(proxmox):
    """
    Asks the user questions about the Proxmox VM to provision.

    Parameters
    ----------
    proxmox: ProxmoxClient

    Returns
    -------
    dict of key-value pairs of answered questions.
    """
    node_q = EnumQuestion("Proxmox Node to create VM on",
                          valid_answers=proxmox.get_nodes())
    node_q.ask()
    chosen_node = node_q.answer

    storage_q = EnumQuestion("Storage to create disk on",
                             valid_answers=proxmox.get_storage(chosen_node))
    storage_q.ask()
    chosen_storage = storage_q.answer

    proxmox_questions = QuestionGroup([
        ("node", NoAskQuestion(question=None, default=chosen_node)),
        ("storage", NoAskQuestion(question=None, default=chosen_storage)),
        ("cpu", IntegerQuestion(
            "Amount of CPUs", min_value=1,
            max_value=proxmox.get_max_cpu(chosen_node))),
        ("cpu_family", EnumQuestion(
            "Emulate which CPU family",
            default="host", valid_answers=CPU_FAMILIES)),
        ("memory", IntegerQuestion(
            "Amount of Memory (MB)", min_value=32,
            max_value=proxmox.get_max_memory(chosen_node))),
        ("disk", IntegerQuestion(
            "Size of disk (GB)", min_value=4,
            max_value=proxmox.get_max_disk_size(chosen_node, chosen_storage))),
        ("vmid", IntegerQuestion("Virtual Machine id", min_value=1,
                                 default=proxmox.get_next_vmid()))
    ])

    proxmox_questions.ask_all()
    return proxmox_questions.flatten_answers()


class ProxmoxClient(object):
    """
    Wrapper around Proxmoxer, to encapsulate retrieval logic in one place.
    """
    def __init__(self, client):
        """
        Parameters
        ----------
        client: ProxmoxAPI
            ProxmoxAPI intance
        """
        self.client = client

    def get_next_vmid(self):
        """
        Retrieve the next available vmid.

        Returns
        -------
        The next available vmid.
        """
        return self.client.cluster.nextid.get()

    def get_nodes(self):
        """
        Retrieve a list of available nodes.

        Returns
        -------
        List of node names.
        """
        return [_node['node'] for _node in self.client.nodes.get()]

    def get_max_cpu(self, node=None):
        """
        Get maximum available cpus.

        Parameters
        ----------
        node: str
            If provided, will retrieve the cpu limit for this specific node. If
            not, will return the lowest common denomitor for all nodes.

        Returns
        -------
        Amount of cpus available.
        """
        if node:
            status = self.client.nodes(node).status.get()
            return status['cpuinfo']['cpus'] * status['cpuinfo']['sockets']
        else:
            return min([_node['maxcpu'] for _node in self.client.nodes.get()])

    def get_max_memory(self, node=None):
        """
        Get maximum amount of memory available.

        Parameters
        ----------
        node: str
            If provided, will retrieve the memory limit for this specific node.
            If not, will return the lowest common denomitor for all nodes.

        Returns
        -------
        Amount of memory available in megabytes.
        """
        if node:
            status = self.client.nodes(node).status.get()
            return int(math.floor(
                status['memory']['total'] / 1024 ** 2))
        else:
            return min(
                [int(math.floor(_node['maxmem'] / 1024 ** 2))
                 for _node in self.client.nodes.get()]
            )

    def get_storage(self, node):
        """
        Get available storages.

        Parameters
        ----------
        node: str
            If provided, will retrieve the storages for this specific node. If
            not, will retrieve the global storages.

        Returns
        -------
        List of storages available.
        """
        storages = []
        for storage in self.client.nodes(node).storage.get():
            if ("images" in storage['content'].split(",")
                    and storage['type'] in ("dir", "lvm")):
                storages.append(storage['storage'])
        return storages

    def get_max_disk_size(self, node=None, storage=None):
        """
        Get the maximum amount of disk space available.

        Parameters
        ----------
        node: str
            If provided, will retrieve the disk limit for this specific node. A
            storage must also be specified. If not, will return the lowest
            common denomitor for all storages.
        storage: str
            Name of storage to lookup maximum amount for.

        Returns
        -------
        Amount of disk space available in gigabytes.
        """
        if node:
            if not storage:
                raise ValueError(
                    "A storage must also be specified for the given node")
            _storage = self.client.nodes(node).storage.get(storage=storage)[0]
            return int(math.floor(
                _storage['avail'] / 1024 ** 3))
        else:
            return min([int(math.floor(_node['maxdisk'] / 1024 ** 3))
                        for _node in self.client.nodes.get()])

    def create_vm(self, node, vmid, name, cpu, cpu_family, memory,
                  vlan_id=None):
        """
        Creates a VM.

        Parameters
        ----------
        node: str
            Name of the node to create the VM on.
        vmid: int
            ID of the VM.
        name: str
            Name of the VM.
        cpu: int
            Number of CPU cores.
        cpu_family: str
            What CPU family to emulate.
        memory: int
            Megabytes of memory.
        vlan_id: int
            VLAN ID of the network device.
        """
        node = self.client.nodes(node)
        net0 = "virtio,bridge=vmbr0"
        if vlan_id:
            net0 += ",tag={0}".format(vlan_id)

        logger.info("Creating Virtual Machine")
        node.qemu.create(
            vmid=vmid, name=name, sockets=1, cores=cpu, cpu=cpu_family,
            memory=memory, net0=net0
        )

    def _upload(self, ssh, filename):
        logger.info("Transferring image to Proxmox")
        tmpfile = os.path.join("/tmp", os.path.basename(filename))
        with open(filename) as _file:
            ssh.upload_file_obj(_file, tmpfile)
        return tmpfile

    def _decompress_image(self, ssh, tmpfile):
        _, ext = os.path.splitext(tmpfile)
        if ext in VALID_COMPRESSION_FORMATS:
            logger.info("Decompressing image")
            if ext == ".xz":
                command = "unxz"
            elif ext == ".gz":
                command = "gunzip"
            else:
                command = "bunzip2"
            stdout, stderr = ssh._exec("{0} '{1}'".format(command,
                                                          tmpfile))
            if len(stdout) > 0 or len(stderr) > 0:
                raise SSHCommandInvocationException(
                    "Failed to decompress image", stdout=stdout, stderr=stderr)
            tmpfile, _ = os.path.splitext(tmpfile)

        _, ext = os.path.splitext(tmpfile)
        if ext not in VALID_IMAGE_FORMATS:
            raise RuntimeError("Provided image is not of a valid type: {0}"
                               .format(", ".join(VALID_IMAGE_FORMATS)))

        return tmpfile

    def _get_virtual_disk_size(self, ssh, tmpfile):
        stdout, stderr = ssh._exec("qemu-img info '{0}'".format(tmpfile))

        if len(stdout) == 0 or len(stderr) > 0:
            raise SSHCommandInvocationException(
                "Failed to get virtual disk size", stdout=stdout,
                stderr=stderr)

        virtual_size = 0
        try:
            for line in stdout.split("\n"):
                if "virtual size" in line:
                    virtual_size = line.split("(")[1].split()[0]
                    virtual_size = int(math.ceil(int(virtual_size) / 1024))
                    break
        except:
            pass
        return virtual_size

    def _allocate_disk(self, ssh, storage, vmid, diskname, disk_size,
                       storagename, disk_format):
        logger.info("Allocating virtual disk")
        stdout, stderr = ssh._exec(
            "pvesm alloc '{0}' {1} '{2}' {3} -format {4}".format(
                storage, vmid, diskname,
                disk_size, disk_format
            )
        )
        if storagename not in stdout and len(stderr) > 0:
            raise SSHCommandInvocationException(
                "Failed to allocate disk", stdout=stdout, stderr=stderr)

    def _get_device_path(self, ssh, storagename):
        stdout, stderr = ssh._exec(
            "pvesm path '{0}'".format(storagename)
        )

        if len(stderr) > 0:
            raise SSHCommandInvocationException(
                "Failed to get path for disk", stdout=stdout, stderr=stderr)

        return stdout.strip()

    def _copy_image_into_disk(self, ssh, disk_format, tmpfile, devicepath):
        logger.info("Copying image into virtual disk")
        stdout, stderr = ssh._exec(
            "qemu-img convert -O {0} '{1}' {2}".format(disk_format, tmpfile,
                                                       devicepath)
        )

        if len(stderr) > 0:
            raise SSHCommandInvocationException(
                "Failed to copy file into disk", stdout=stdout, stderr=stderr)

    def _upload_to_storage(self, ssh_session, storage, vmid, filename,
                           diskname, storagename, disk_format="raw",
                           disk_size=None):
        """
        Upload a file into a datastore. The steps executed are:
          1. The file is uploaded via SFTP to /tmp.
          2. A new disk is allocated using `pvesm`.
          3. The path of this disk is retrieved using `pvesm`.
          4. The file is converted and transfered into the disk
          using `qemu-img`.
          5. The temporary file is removed.

        Parameters
        ----------
        ssh_session: ProxmoxBaseSSHSession subclass
            This is an internal class used by proxmoxer, we're using its ssh
            transfer methods for various operations.
        storage: str
            Name of storage to upload the file into.
        vmid: int
            ID of the VM to associate the file with. This is enforced by
            Proxmox.
        filename: str
            Local filename of the file.
        diskname: str
            Name of the disk to allocate.
        storagename: str
            Full canonical name of the disk.
        disk_format: raw or qcow2
            Format of the file. The source type doesn't matter, as we will call
            `qemu-img` to both transfer and convert the file into the disk.
        disk_size: int
            Override the disk size. If not specified, the size is calculated
            from the file. In kilobytes.
        """
        try:
            tmpfile = self._upload(ssh_session, filename)
            tmpfile = self._decompress_image(ssh_session, tmpfile)
            image_size = self._get_virtual_disk_size(ssh_session, tmpfile)

            if not disk_size:
                logger.warning("Setting disk size to {0}K".format(image_size))
                disk_size = image_size
            elif image_size > disk_size:
                logger.warning("Provided disk size was too small, "
                               "increasing to {0}K".format(image_size))
                disk_size = image_size

            self._allocate_disk(ssh_session, storage, vmid, diskname,
                                disk_size, storagename, disk_format)

            devicepath = self._get_device_path(ssh_session, storagename)

            self._copy_image_into_disk(ssh_session, disk_format, tmpfile,
                                       devicepath)
        finally:
            if tmpfile:
                logger.info("Removing temporary disk file")
                ssh_session._exec("rm '{0}'".format(tmpfile))

    def _upload_to_flat_storage(self, storage, vmid, filename, disk_format,
                                disk_label, disk_size=None):
        """
        Generates appropriate names for uploading a file to a 'dir' datastore.
        Actual work is done by _upload_to_storage.

        Parameters
        -----------
        storage: str
            Name of storage to upload the file into.
        vmid: int
            ID of the VM to associate the file with. This is enforced by
            Proxmox.
        filename: str
            Local filename of the file.
        disk_format: raw or qcow2
            Format of the file. The source type doesn't matter, as we will call
            `qemu-img` to both transfer and convert the file into the disk.
        disk_label: str
            Label to incorporate in the resulting disk name.
        disk_size: int
            Override the disk size. If not specified, the size is calculated
            from the file. In kilobytes.

        Returns
        -------
        Full canonical name of the disk.
        """
        ssh_session = self.client._backend.session
        diskname = "vm-{0}-{1}.{2}".format(vmid, disk_label, disk_format)
        storagename = "{0}:{1}/{2}".format(storage, vmid, diskname)

        logger.info("Uploading to flat storage")
        self._upload_to_storage(ssh_session, storage, vmid, filename,
                                diskname, storagename, disk_format=disk_format,
                                disk_size=disk_size)

        return storagename

    def _upload_to_lvm_storage(self, storage, vmid, filename, disk_format,
                               disk_label, disk_size=None):
        """
        Generates appropriate names for uploading a file to a 'lvm' datastore.
        Actual work is done by _upload_to_storage.

        Parameters
        -----------
        storage: str
            Name of storage to upload the file into.
        vmid: int
            ID of the VM to associate the file with. This is enforced by
            Proxmox.
        filename: str
            Local filename of the file.
        disk_format: raw or qcow2
            Format of the file. Will be overridden into 'raw', because LVM only
            supports RAW disks.
        disk_label: str
            Label to incorporate in the resulting disk name.
        disk_size: int
            Override the disk size. If not specified, the size is calculated
            from the file. In kilobytes.

        Returns
        -------
        Full canonical name of the disk.
        """
        ssh_session = self.client._backend.session
        diskname = "vm-{0}-{1}".format(vmid, disk_label)
        storagename = "{0}:{1}".format(storage, diskname)

        logger.info("Uploading to LVM storage")
        # LVM only supports raw disks, overwrite the disk_format here.
        self._upload_to_storage(ssh_session, storage, vmid, filename,
                                diskname, storagename, disk_format="raw",
                                disk_size=disk_size)

        return storagename

    def upload(self, node, storage, vmid, filename, disk_format, disk_label,
               disk_size=None):
        """
        Upload a file into a datastore.

        Note that we can't yet upload a file to another node, only to the local
        node that we have an SSH connection with.

        Parameters
        ----------
        node: str
            Name of the node to upload to. See the note above.
        storage: str
            Name of storage to upload the file into.
        vmid: int
            ID of the VM to associate the file with. This is enforced by
            Proxmox.
        filename: str
            Local filename of the file.
        disk_format: raw or qcow2
            Format of the file.
        disk_label: str
            Label to incorporate in the resulting disk name.
        disk_size: int
            Override the disk size. If not specified, the size is calculated
            from the file. In kilobytes.
        """
        _node = self.client.nodes(node)
        _storage = _node.storage(storage)
        _type = _storage.status.get()['type']
        if _type == "dir":
            diskname = self._upload_to_flat_storage(
                storage=storage, vmid=vmid, filename=filename,
                disk_label=disk_label, disk_format=disk_format,
                disk_size=disk_size)
        elif _type == "lvm":
            diskname = self._upload_to_lvm_storage(
                storage=storage, vmid=vmid, filename=filename,
                disk_label=disk_label, disk_format=disk_format,
                disk_size=disk_size)
        else:
            raise ValueError(
                "Only dir and lvm storage are supported at this time")
        return diskname

    def attach_seed_iso(self, node, storage, vmid, iso_file):
        """
        Upload a cloud-init seed ISO file, and attach it to a VM.

        Parameters
        ----------
        node: str
            Name of the node to upload to. See the note above.
        storage: str
            Name of storage to upload the file into.
        vmid: int
            ID of the VM to associate the file with. This is enforced by
            Proxmox.
        iso_file: str
            Local filename of the ISO file.
        """
        _node = self.client.nodes(node)
        diskname = self.upload(node, storage, vmid, iso_file,
                               disk_label="cloudinit-seed", disk_format="raw")
        _node.qemu(vmid).config.set(virtio1=diskname)

    def attach_base_disk(self, node, storage, vmid, img_file, disk_size):
        """
        Upload a Cloud base image, and attach it to a VM.

        Parameters
        ----------
        node: str
            Name of the node to upload to. See the note above.
        storage: str
            Name of storage to upload the file into.
        vmid: int
            ID of the VM to associate the file with. This is enforced by
            Proxmox.
        img_file: str
            Local filename of the ISO file.
        disk_size: int
            Size of the disk to allocate, in kilobytes.
        """
        _node = self.client.nodes(node)
        diskname = self.upload(node, storage, vmid, img_file,
                               disk_label="base-disk", disk_format="qcow2",
                               disk_size=disk_size)
        _node.qemu(vmid).config.set(virtio0=diskname, bootdisk="virtio0")
        try:
            logger.info("Resizing virtual disk")
            _node.qemu(vmid).resize.set(disk="virtio0", size=disk_size * 1024)
        except SSHError as se:
            if "disk size" not in str(se):
                raise se
            logger.error("Failed to set disk size, disk will probably be "
                         "bigger than expected")



# proxmox-deploy is cli-based deployment tool for Proxmox
#
# Copyright (c) 2015 Nick Douma <n.douma@nekoconeko.nl>
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see http://www.gnu.org/licenses/.



# proxmox-deploy is cli-based deployment tool for Proxmox
#
# Copyright (c) 2015 Nick Douma <n.douma@nekoconeko.nl>
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see http://www.gnu.org/licenses/.

from collections import OrderedDict
from contextlib import contextmanager
import sys


class QuestionGroup(OrderedDict):
    """
    A group of Questions, usually about the same subject.
    """
    def __init__(self, questions, *args, **kwargs):
        """
        Parameters
        ----------
        questions: list of tuples
            List of (key, question) tuples. Question order is preserved.
        """
        super(QuestionGroup, self).__init__(*args, **kwargs)
        for key, question in questions:
            self[key] = question

    def ask_all(self, _output=None, _input=None):
        """
        Call the ask() method on all Questions. If nested QuestionGroups are
        encountered, ask_all() on those as well.

        Parameters
        ----------
        _output: file
            Allows for temporarily overriding the output file on each Question.
        _input: file
            Allows for temporarily overriding the input file on each Question.
        """
        for question in self.values():
            if isinstance(question, QuestionGroup):
                question.ask_all()
            else:
                question.ask(_output=_output, _input=_input)

    def flatten_answers(self):
        """
        Flattens all Question answers into a dictionary. Nested QuestionGroups
        are also flattened.
        """
        answers = {}
        for key, question in self.iteritems():
            if isinstance(question, QuestionGroup):
                _answers = question.flatten_answers()
                answers.update(_answers)
            else:
                answers[key] = question.answer
        return answers

    def lookup_answer(self, key):
        return self.flatten_answers()[key]


class OptionalQuestionGroup(QuestionGroup):
    def __init__(self, questions, optional_question, negative_questions=None,
                 *args, **kwargs):
        super(OptionalQuestionGroup, self).__init__(questions, *args, **kwargs)
        self.optional_question = optional_question
        self.negative_questions = negative_questions

    def evaluate_answer(self):
        return bool(self.optional_question.answer)

    def ask_all(self, _output=None, _input=None):
        self.optional_question.ask(_output=_output, _input=_input)
        if self.evaluate_answer():
            super(OptionalQuestionGroup, self)\
                .ask_all(_output=_output, _input=_input)

    def flatten_answers(self):
        if self.evaluate_answer():
            return super(OptionalQuestionGroup, self).flatten_answers()
        elif self.negative_questions:
            return self.negative_questions
        else:
            return {}


class SpecificAnswerOptionalQuestionGroup(OptionalQuestionGroup):
    def __init__(self, questions, optional_question, specific_answer,
                 *args, **kwargs):
        super(SpecificAnswerOptionalQuestionGroup, self).__init__(
            questions, optional_question, *args, **kwargs
        )
        self.specific_answer = specific_answer

    def evaluate_answer(self):
        return self.specific_answer == self.optional_question.answer


class Question(object):
    """
    Base Question class, which accepts all answers and stores them as string.
    """
    question_without_default = "{0}: "
    question_with_default = "{0} [{1}]: "

    def __init__(self, question, default=None,
                 _output=sys.stderr, _input=sys.stdin):
        """
        Asks an interactive question and stores the answer. See ask() for the
        workflow.

        Parameters
        ----------
        question: str
            Question to ask the user. Will be formatted and written to output.
        default: any
            Default value. Will be formatted as string and outputted as part of
            the question.
        _output: file
            File to output questions to. sys.stderr by default.
        _input: file
            File to read input from. sys.stdin by default.
        """
        self.question = question
        self.answer = default
        self.input = _input
        self.output = _output

    def _format_question(self):
        if self.answer is not None:
            return self.question_with_default.format(
                self.question,
                self.format_default()
            )
        else:
            return self.question_without_default.format(self.question)

    def _write_question(self):
        self.output.write(self._format_question())

    def _read_answer(self):
        return self.input.readline().strip()

    @contextmanager
    def _override_files(self, _output, _input):
        old_output = self.output
        old_input = self.input
        if _output:
            self.output = _output
        if _input:
            self.input = _input
        yield
        self.output = old_output
        self.input = old_input

    def ask(self, _output=None, _input=None):
        """
        Asks the question to the user. The internal workflow of method calls
        is:
            1. Print question (_write_question and format_default).
            2. Read answer (_read_answer).
            3. If answer was empty and a default was set, use default.
            4. Else, validate the answer (validate).
            5. If valid, format answer (format_answer).
            6. Else, go to 1.

        format_default, format_answer and validate are implemented in the base
        class, but should probably be overridden in subclasses.
        """
        with self._override_files(_output, _input):
            valid = False
            while not valid:
                self._write_question()
                answer = self._read_answer()
                if answer == "" and self.answer is not None:
                    return
                valid = self.validate(answer)
            self.answer = self.format_answer(answer)

    def format_default(self):
        """
        Formats the default value for output to user. In the base class, the
        default will simply be cast to string.
        """
        return str(self.answer)

    def format_answer(self, answer):
        """
        Formats the answer before storing it. In the base class, the answer
        is returned unmodified.
        """
        return answer

    def validate(self, answer):
        """
        Validates the given answer. In the base class, the answer is valid when
        it is not empty.

        This method is responsible for outputting helpful messages if the
        answer is invalid.

        Should return False when invalid.
        """
        return answer != ""


class BooleanQuestion(Question):
    """
    Question class which only accepts boolean answers.
    """
    positive_answers = ["true", "t", "yes", "y"]
    negative_answers = ["false", "f", "no", "n"]

    def validate(self, answer):
        """
        Validates if the given answer is contained in positive_answers or
        negative_answers (case-insensitive).
        """
        if answer.lower() in self.positive_answers:
            answer = True
            return True
        elif answer.lower() in self.negative_answers:
            answer = False
            return True
        else:
            self.output.write("Please answer 'Yes' or 'No'.\n")
            return False

    def format_default(self):
        """
        Formats the default into a user friendly "Yes" or "No".
        """
        if self.answer is True:
            return "Yes"
        else:
            return "No"

    def format_answer(self, answer):
        """
        Converts the given answer into a boolean for storage.
        """
        if answer.lower() in self.positive_answers:
            return True
        else:
            return False


class IntegerQuestion(Question):
    """
    Question class which only accepts integer answers.
    """
    def __init__(self, question, min_value=None, max_value=None, **kwargs):
        super(IntegerQuestion, self).__init__(question, **kwargs)
        self.min_value = min_value
        self.max_value = max_value

    def validate(self, answer):
        """
        Validates the given answer by casting it to an integer. If a min_value
        and/or max_value was supplied to the constructor, also validates the
        answer against that.
        """
        try:
            _answer = int(answer)
        except ValueError:
            if not self.min_value and not self.max_value:
                self.output.write("Please enter a valid integer.\n")
            elif self.min_value and not self.max_value:
                self.output.write(
                    "Please enter a valid integer bigger than {0}.\n"
                    .format(str(self.min_value))
                )
            elif not self.min_value and self.max_value:
                self.output.write(
                    "Please enter a valid integer smaller than {0}.\n"
                    .format(str(self.max_value))
                )
            else:
                self.output.write(
                    "Please enter a valid integer between {0} and {1}.\n"
                    .format(str(self.min_value), str(self.max_value))
                )
            return False

        if self.min_value and _answer < self.min_value:
            self.output.write(
                "Please enter a value bigger than {0}.\n"
                .format(str(self.min_value))
            )
            return False

        if self.max_value and _answer > self.max_value:
            self.output.write(
                "Please enter a value smaller than {0}.\n"
                .format(str(self.max_value))
            )
            return False
        return True

    def format_answer(self, answer):
        """
        Casts the answer to an integer.
        """
        return int(answer)


class EnumQuestion(Question):
    """
    Question class which only accepts answer from a given list.
    """
    question_without_default = "{0} (Enter ? for a list of options): "
    question_with_default = "{0} (Enter ? for a list of options) [{1}]: "

    def __init__(self, question, valid_answers, default=None, **kwargs):
        super(EnumQuestion, self).__init__(question, default, **kwargs)
        assert len(valid_answers) > 0
        if default:
            assert default in valid_answers
        self.valid_answers = valid_answers

    def validate(self, answer):
        """
        Validates the given answer by checking it's presence in the
        provided list.
        """
        if answer == "?" or answer not in self.valid_answers:
            self.output.write(
                "Please enter one of: {0}\n"
                .format(", ".join(self.valid_answers))
            )
            return False
        return True


class MultipleAnswerQuestion(Question):
    """
    Question class which accepts multiple answers.
    """
    def ask(self, _output=None, _input=None):
        """
        Asks the question to the user, like Question. After asking the question
        once, the user is prompted to enter more values until an empty value is
        provided.
        """
        with self._override_files(_output, _input):
            answers = []
            defaults = self.answer
            valid = False
            while not valid:
                self._write_question()
                answer = self._read_answer()
                if answer and defaults:
                    defaults = None
                if answer == "" and self.answer is not None:
                    return
                valid = self.validate(answer)
            answers.append(answer)

            while answer:
                answer = None
                self.output.write(
                    "Enter another value, or empty to continue: ")
                answer = self._read_answer()
                if not answer:
                    self.answer = answers
                    return
                else:
                    valid = self.validate(answer)
                    if valid:
                        answers.append(answer)

    def format_default(self):
        """
        Formats the default value for output to user. Only the count of
        elements in the answer is returned.
        """
        return "{0} entries".format(len(self.answer))


class NoAskQuestion(Question):
    """
    Question class which only supplies an answer without asking a Question.
    """
    def __init__(self, question, default, **kwargs):
        super(NoAskQuestion, self).__init__(question, default, **kwargs)

    def ask(self, _output=None, _input=None):
        """
        No questions asked.
        """
        pass

    def format_answer(self, answer):
        """
        Return the answer without any casting or converting.
        """
        return answer



# proxmox-deploy is cli-based deployment tool for Proxmox
#
# Copyright (c) 2015 Nick Douma <n.douma@nekoconeko.nl>
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see http://www.gnu.org/licenses/.

from setuptools import setup, find_packages
from proxmoxdeploy.version import NAME, VERSION, DESCRIPTION

with open('README.rst') as file:
    long_description = file.read()

setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=long_description,
    author="Nick Douma",
    author_email="n.douma@nekoconeko.nl",
    url="https://github.com/LordGaav/proxmox-deploy",
    license="GPLv3",
    packages=find_packages(),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Operating System :: POSIX :: Linux',
        'Operating System :: POSIX :: BSD',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.0',
        'Programming Language :: Python :: 3.1',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python',
        'Topic :: System :: Systems Administration',
        'Topic :: Utilities'
    ],
    install_requires=[
        'proxmoxer',
        'openssh-wrapper',
        'Jinja2',
        'configobj',
        'pytz'
    ],
    package_data={"": ["*.j2"]},
    entry_points={
        'console_scripts': [
            'proxmox-deploy = proxmoxdeploy.cli:main'
        ]
    }
)



# proxmox-deploy is cli-based deployment tool for Proxmox
#
# Copyright (c) 2015 Nick Douma <n.douma@nekoconeko.nl>
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see http://www.gnu.org/licenses/.

from .cloudinit.templates import ask_cloudinit_questions
from .cloudinit import generate_seed_iso
from .exceptions import CommandInvocationException
from .proxmox import ProxmoxClient, ask_proxmox_questions
from .version import NAME, VERSION, BUILD, DESCRIPTION
from argparse import ArgumentParser
from configobj import ConfigObj
from proxmoxer import ProxmoxAPI, ResourceException
import logging
import os
import sys

root_logger = logging.getLogger(None)
root_logger.addHandler(logging.StreamHandler())
base_logger = logging.getLogger("proxmoxdeploy")
base_logger.setLevel(logging.INFO)
logger = logging.getLogger("proxmoxdeploy.cli")


def get_arguments():
    initial_parser = ArgumentParser(add_help=False)
    initial_parser.add_argument("--config", metavar="CFG", type=str,
                                help="Config file to load")
    initial_parser.add_argument("--version", action="store_true",
                                default=False,
                                help="Display version information and exit")

    args, unknown_args = initial_parser.parse_known_args()

    config = {}
    if args.config:
        config = ConfigObj(args.config)

    parser = ArgumentParser(description=DESCRIPTION)
    parser.add_argument("--config", metavar="CFG", type=str,
                        help="Config file to load.")
    parser.add_argument("--proxmox-host", metavar="HOST", type=str,
                        default=config.get("proxmox-host", None),
                        help="Proxmox API host.")
    parser.add_argument("--proxmox-user", metavar="USER", type=str,
                        default=config.get("proxmox-user", "root"),
                        help="Proxmox API user.")
    parser.add_argument("--cloud-images-dir", metavar="DIR", type=str,
                        default=config.get("cloud-images-dir", None),
                        help="Directory containing Cloud images.")
    args = parser.parse_args()

    if not args.proxmox_host:
        logger.error("No Proxmox API host was supplied.")
        sys.exit(1)

    if not args.cloud_images_dir:
        logger.error("No directory containing Cloud images specified.")
        sys.exit(1)

    return args


def interact_with_user(args, api):
    proxmox_answers = ask_proxmox_questions(api)
    cloudinit_answers = ask_cloudinit_questions(
        cloud_images_dir=args.cloud_images_dir)
    return (proxmox_answers, cloudinit_answers)


def main():
    logger.info("{0} version {1} (build {2}) starting...".format(
        NAME, VERSION, BUILD))

    args = get_arguments()
    api = ProxmoxClient(ProxmoxAPI(args.proxmox_host, port="22", timeout=600,
                                   user=args.proxmox_user, backend="openssh"))

    logger.info("Asking user for configuration input")
    try:
        (proxmox, cloudinit) = interact_with_user(args, api)
        context = dict(proxmox, **cloudinit)
    except KeyboardInterrupt:
        logger.info("Aborted by user")
        sys.exit(0)

    logger.info("")
    logger.info("")
    logger.info("Starting provisioning process")

    try:
        api.create_vm(node=proxmox['node'], vmid=proxmox['vmid'],
                      name=cloudinit['name'], cpu=proxmox['cpu'],
                      cpu_family=proxmox['cpu_family'],
                      memory=proxmox['memory'], vlan_id=cloudinit['vlan_id'])
    except ResourceException:
        logger.error("Failed to create VM")
        sys.exit(1)

    try:
        cloudinit_iso = generate_seed_iso(context=context)
        logger.debug("File generated at: {0}".format(cloudinit_iso))

        logger.info("Uploading cloud-init seed ISO to Proxmox")
        api.attach_seed_iso(node=proxmox['node'], storage=proxmox["storage"],
                            vmid=proxmox['vmid'], iso_file=cloudinit_iso)
        logger.info("Uploading cloud image to Proxmox")
        disk_size = proxmox['disk'] * 1024 ** 2
        api.attach_base_disk(node=proxmox['node'], storage=proxmox["storage"],
                             vmid=proxmox['vmid'],
                             img_file=cloudinit['image'],
                             disk_size=disk_size)
    except CommandInvocationException as cie:
        logger.error("Provisioning failed")
        if hasattr(cie, "stdout") or hasattr(cie, "stderr"):
            logger.error("Command output was:")
        if hasattr(cie, "stdout"):
            logger.error(cie.stdout)
        if hasattr(cie, "stderr"):
            logger.error(cie.stderr)
        sys.exit(1)
    finally:
        if os.path.exists(cloudinit_iso):
            logger.debug("Removing seed ISO file")
            os.remove(cloudinit_iso)

    logger.info("Virtual Machine provisioning completed")

if __name__ == "__main__":
    main()
